//----------------------------------------------------------------------------------------------
// BSP : Seeed nRF52 Borads 1.1.9
// Board : Seeed nRF52 Borads / Seeed XIAO nRF52840
//----------------------------------------------------------------------------------------------
// Peripheral sketch : nRF52_XIAO_XBLE_range_test_peripheral.ino
// 2025/03/31

#include <Wire.h>
#include <Adafruit_GFX.h>
#include <Adafruit_SSD1306.h>
#include "myFonts/tahoma7pt7b.h"
#include <bluefruit.h>

#define VER   "20250331c"   // sketch version
#define DATANUM       244   // max MTU = 247, DATANUM = MTU - 3, See ble_gap.h:253

// Instances of the class
Adafruit_SSD1306 display(128, 64, &Wire, -1);   // oled display SSD1306

union unionData {                   // Union for data type conversion
  float     dataBuffFL[DATANUM/4];
  uint32_t  dataBuff32[DATANUM/4];
  uint16_t  dataBuff16[DATANUM/2];
  uint8_t   dataBuff8[DATANUM];
};
union unionData ud;
uint8_t readbuff[DATANUM];          // for packet read
int avebuff[16];                    // for moveing average
char* versionNumber = VER;          // central sketch version number
bool receivedFlag = false;          // central received flag
bool connectedFlag = false;         // connect flag
int8_t txpower = -127;               // Peripheral Tx power -20 ~ 8dBm, 4dBm step    

// Custum Client Service and Characteristic
// 55c40000-xxxx-4dd1-be0c-40588193b485 for example
#define myUUID(val) (const uint8_t[]) { \
    0x85, 0xB4, 0x93, 0x81, 0x58, 0x40, 0x0C, 0xBE, \
    0xD1, 0x4D, (uint8_t)(val & 0xff), (uint8_t)(val >> 8), 0x00, 0x00, 0xC4, 0x55 }
BLEClientService        P2C2P_Service      (myUUID(0x0000)); // peripheral -> central -> peripheral
BLEClientCharacteristic P2C_Characteristic (myUUID(0x0030)); // peripheral -> central
BLEClientCharacteristic C2P_Characteristic (myUUID(0x0100)); // central -> peripheral

void setup()
{
  // Serial port initialization
  Serial.begin(115200);
//  while (!Serial);
  delay(1000);

  // LED intialization
  pinMode(LED_RED, OUTPUT);
  pinMode(LED_GREEN, OUTPUT);
  pinMode(LED_BLUE, OUTPUT);

  digitalWrite(LED_RED, HIGH);
  digitalWrite(LED_GREEN, HIGH);
  digitalWrite(LED_BLUE, HIGH);

  // Initialization of display SSD1306
  Wire.begin();
  if (!display.begin(SSD1306_SWITCHCAPVCC, 0x3C))
  {
    Serial.println("Initialization failure of SSD1306");
  }
  display.clearDisplay();
  display.setFont(&tahoma7pt7b);
  display.setTextColor(SSD1306_WHITE);
  display.setCursor(0, 15);
  display.println("Extended BLE");
  display.setCursor(0, 31);
  display.println("    Range Test");
  display.setCursor(0, 47);
  display.println("        Central");  
  display.setCursor(0, 63);
  display.print("Ver   : ");
  display.println(versionNumber);
  display.display();
  delay(5000);
  display.clearDisplay();

  // Initialization of Bruefruit
  Bluefruit.configCentralBandwidth(BANDWIDTH_MAX);  
  Bluefruit.begin(0, 1);    // central
  Bluefruit.setTxPower(8);  // ******************************************* default
  Bluefruit.setName("XIAO BLE Central");
//  Bluefruit.setConnLedInterval(100);
  Bluefruit.autoConnLed(false);   // BLE not used Blue LED
  Bluefruit.Central.setDisconnectCallback(disconnect_callback);
  Bluefruit.Central.setConnectCallback(connect_callback);

  // Custom Service Settings
  P2C2P_Service.begin();
//  P2C_Characteristic.setIndicateCallback(data_received_callback);
  P2C_Characteristic.setNotifyCallback(data_received_callback);
  P2C_Characteristic.begin(); 
  C2P_Characteristic.begin();

  // Scanner settings
  Bluefruit.Scanner.setRxCallback(scan_callback);
  Bluefruit.Scanner.restartOnDisconnect(true);
  Bluefruit.Scanner.setIntervalMS(100, 50);     // interval:100mS  window:50mS  
  Bluefruit.Scanner.filterUuid(P2C2P_Service.uuid);
  Bluefruit.Scanner.useActiveScan(false);
  Bluefruit.Scanner.start(0);                   // Continuous Scanning

  Serial.println("End of initialization");

  // Wait for response from peripheral
  Serial.println("Connecting to Peripheral ....");
  while (!Bluefruit.connected()) {
    Serial.print("."); delay(100);
  }
  Serial.println();
  Serial.println("Connected to Peripheral");
}

//******************************************************************************************
void loop()
{
  if(connectedFlag == true)
  {
    digitalWrite(LED_BLUE, LOW);

    if(receivedFlag == true)
    {
      // receive a packet from peripheral --------------------------------------------------
      digitalWrite(LED_GREEN, LOW);
        // receive a packet      
        uint32_t dataLength = sizeof(readbuff);
        memcpy(ud.dataBuff8, readbuff, dataLength);
        receivedFlag = false;

        // get peripheral parameters
        uint32_t ms = ud.dataBuff32[0];            // time transmitted by peripheral        
        uint32_t verifynumber = ud.dataBuff32[4];  // random number transmitted by peripheral

        uint16_t conn_hdl = P2C_Characteristic.connHandle();
        BLEConnection* connection = Bluefruit.Connection(conn_hdl);
        uint8_t phy = connection->getPHY();
        int8_t rssi = connection->getRssi();
          // rssi moveing average
          int avenum = 10;
          for(int i = 0; i < avenum; i++) {
            avebuff[i] = avebuff[i + 1];
          }
          avebuff[avenum - 1] = rssi;
          int32_t sumbuff = 0;
          for(int i = 0; i < avenum; i++) {
            sumbuff = sumbuff + avebuff[i];
          }
          float ave_rssi = (float)sumbuff / avenum;

        // print and display
        Serial.print("RSSI = "); Serial.println(ave_rssi, 1);
        Serial.print("dataLength = "); Serial.println(dataLength);
        Serial.print("PHY = "); Serial.println(phy);
        Serial.print("mS = "); Serial.println(ms);
        Serial.print("Txpower = "); Serial.println(txpower);
        Serial.printf("VerifyNumber = %08x\n", verifynumber);       
        Serial.println("**************************************");

        display.clearDisplay();
        display.setCursor(0, 15); display.print(verifynumber, HEX); 
        display.setCursor(80, 31);display.print(ms / 1000); display.print(" s");
        display.setCursor(0, 47); display.print(dataLength); display.print(" byte");
        display.setCursor(64, 47);
          if      (phy == 1) { display.print("1 Mbps"); } 
          else if (phy == 2) { display.print("2 Mbps"); } 
          else if (phy == 4) { display.print("Coded"); } 
          else               { display.print("error"); }
        display.setCursor(0, 63); display.print(txpower); display.print(" dBm");
        display.setCursor(64, 63); display.print(ave_rssi, 1); display.print(" dBm");
        display.display();
      digitalWrite(LED_GREEN, HIGH);

      // transmit a packet to peripheral -------------------------------------------------------------     
      digitalWrite(LED_RED, LOW);
        // setting a backet to peripheral
        ud.dataBuff32[0] = ms;            // test start time
        ud.dataBuff32[1] = verifynumber;  // random number transmitted by peripheral
        ud.dataBuffFL[2] = ave_rssi;      // average rssi of central
        ud.dataBuff32[3] = dataLength;    // data length of packet
        ud.dataBuff8[16] = txpower;       // Peripheral Tx power
        ud.dataBuff8[17] = phy;           // phy, 1:1Mbps 2:2Mbps 4:coded

        C2P_Characteristic.write(ud.dataBuff8, sizeof(ud.dataBuff8));
        delay(10); // for LED
      digitalWrite(LED_RED, HIGH);
    } // received
  } // connect
//  delay(100);
  digitalWrite(LED_RED, HIGH);
  digitalWrite(LED_GREEN, HIGH);
  digitalWrite(LED_BLUE, HIGH);
} // loop

//*****************************************************************************************************
// Callback invoked when scanner pick up an advertising data
void scan_callback(ble_gap_evt_adv_report_t* report)
{
  (void) report;

  Bluefruit.Central.connect(report);

  uint8_t buffer[32];
  memset(buffer, 0, sizeof(buffer));

  // Tx Power Level
  if(Bluefruit.Scanner.parseReportByType(report, BLE_GAP_AD_TYPE_TX_POWER_LEVEL, buffer, sizeof(buffer)))
  {
    txpower = int8_t(buffer[0]);
    memset(buffer, 0, sizeof(buffer));
  }
}

//****************************************************************************************************:
// Callback invoked when an connection is established
void connect_callback(uint16_t conn_handle)
{
  (void) conn_handle;

  connectedFlag = false;
  // scan stop
  Bluefruit.Scanner.stop();
  
  if(!P2C2P_Service.discover(conn_handle)) {
    Serial.println("【connect_callback】 Service not find disconnect!");
    Bluefruit.disconnect(conn_handle);
    return;
  }
  Serial.println("【connect_callback】 Find Service");
  
  if (!P2C_Characteristic.discover()) {
    Serial.println("【connect_callback】 Characteristic not find disconnect!");
    Bluefruit.disconnect(conn_handle);
    return;
  }
  Serial.println("【connect_callback】 Find Characteristic");
  
//  if ( !P2C_Characteristic.enableIndicate() ) {
  if (!P2C_Characteristic.enableNotify()) {
    Serial.println("【connect_callback】 Couldn't enable reception for data");
    Bluefruit.disconnect(conn_handle);
    return;
  }
  Serial.println("【connect_callback】 enable reception");

  if (!C2P_Characteristic.discover()) {
    Serial.println("【connect_callback】 Characteristic not find disconnect!");
    Bluefruit.disconnect(conn_handle);
    return;
  }
  Serial.println("【connect_callback】 Find Characteristic");  

  // start rssi monitor
  BLEConnection* conn = Bluefruit.Connection(conn_handle);
  conn->monitorRssi();

  connectedFlag = true;

  // scan resume
  Bluefruit.Scanner.resume();
}

//**********************************************************************************************************
// Callback invoked when an connection is broken
void disconnect_callback(uint16_t conn_handle, uint8_t reason)
{
  (void) conn_handle; 
  (void) reason;

  Serial.print("【disconnect_callback】 Disconnected, reason = 0x"); Serial.println(reason, HEX);
  connectedFlag = false;
}

//*********************************************************************************************************
// Callback invoked when a packet is received
void data_received_callback(BLEClientCharacteristic* chr, uint8_t* data, uint16_t len)
{
  (void) chr; 
  (void) data; 
  (void) len;

  Serial.print("【received_callback】 len = "); Serial.println(len);
  memcpy(readbuff, data, len);
  receivedFlag = true;
}
