//----------------------------------------------------------------------------------------------
// BSP : Seeed nRF52 Borads 1.1.9
// Board : Seeed nRF52 Boards / Seeed XIAO nRF52840 Sense
//----------------------------------------------------------------------------------------------
// Central sketch : nRF52_XIAO_XBLE_renge_test_central.ino
// 2025/03/31

#include <Wire.h>
#include <Adafruit_GFX.h>
#include <Adafruit_SSD1306.h>
#include "myFonts/tahoma7pt7b.h"
#include <bluefruit.h>
#include <Adafruit_SPIFlash.h>

#define VER     "20250331c"   // sketch version
#define MTU             247   // default 23 -> max247,  See ble_gap.h:253
#define CONNINTERVAL    32    // 40[mS](in unit 1.25) so that the maximum CodedPhy packets fit within the connection interval
#define DATANUM         244   // number of data, MTU - header3
#define INTERVAL       1000   // test interval [mS]

// Instances of the class
Adafruit_SSD1306 display(128, 64, &Wire, -1);   // oled display SSD1306
Adafruit_FlashTransport_QSPI flashTransport;    // on board Flash enter to Deep Power-Down Mode
Adafruit_SPIFlash flash(&flashTransport);

union unionData {                   // Union for data type conversion
  float     dataBuffFL[DATANUM/4];
  uint32_t  dataBuff32[DATANUM/4];
  uint16_t  dataBuff16[DATANUM/2];
  uint8_t   dataBuff8[DATANUM];
};
union unionData ud;
char* versionNumber = VER;          // central sketch version number
uint8_t phy = BLE_GAP_PHY_1MBPS;    // See ble_gap.h:549
int8_t txpower = 8;                 // Tx power -20 ~ 8dBm, 4dBm step
bool writtenFlag = false;           // central written flag
bool connectFlag = false;           // connect flag
uint8_t readbuff[DATANUM];          // for packet read

// Custum Service and Characteristic
// 55c40000-xxxx-4dd1-be0c-40588193b485 for example
#define myUUID(val) (const uint8_t[]) { \
    0x85, 0xB4, 0x93, 0x81, 0x58, 0x40, 0x0C, 0xBE, \
    0xD1, 0x4D, (uint8_t)(val & 0xff), (uint8_t)(val >> 8), 0x00, 0x00, 0xC4, 0x55 }
BLEService        P2C2P_Service       (myUUID(0x0000)); // peripheral -> central -> peripheral
BLECharacteristic P2C_Characteristic  (myUUID(0x0030)); // peripheral -> central
BLECharacteristic C2P_Characteristic  (myUUID(0x0100)); // central -> peripheral

void setup()
{
  Serial.begin(115200);
//  while (!Serial);
  delay(1000);

  // Reduced current consumption
  NRF_POWER->DCDCEN = 1;            // Enable DC/DC converter for REG1 stage
  flashTransport.begin();           // on board Flash enter to Deep Power-Down Mode
  flashTransport.runCommand(0xB9);
  delayMicroseconds(5);
  flashTransport.end();
  delay(1000);
    
  // Pins initialization   
  pinMode(LED_RED, OUTPUT);
  pinMode(LED_GREEN, OUTPUT);
  pinMode(LED_BLUE, OUTPUT);
  pinMode(D0, INPUT_PULLUP);  // for experiment
  pinMode(D3, INPUT_PULLUP);  
  pinMode(D7, INPUT_PULLUP);
  pinMode(D8, INPUT_PULLUP);
  pinMode(D9, INPUT_PULLUP);
  pinMode(D10, INPUT_PULLUP);

  digitalWrite(LED_RED, HIGH);
  digitalWrite(LED_GREEN, HIGH);
  digitalWrite(LED_BLUE, HIGH);

  // phy select for experiment, See ble_gap.h:549
  if      (digitalRead(D0) == LOW)  { phy = BLE_GAP_PHY_1MBPS; } 
  else if (digitalRead(D3) == LOW)  { phy = BLE_GAP_PHY_2MBPS; }
  else                              { phy = BLE_GAP_PHY_CODED; }
  // TxPower select for experiment
  if      (digitalRead(D10) == LOW) { txpower = -8; }
  else if (digitalRead(D9) == LOW)  { txpower = -4; }
  else if (digitalRead(D8) == LOW)  { txpower =  0; }
  else if (digitalRead(D7) == LOW)  { txpower =  4; }
  else                              { txpower =  8; }

  // Initialization of display SSD1306
  Wire.begin();
  if (!display.begin(SSD1306_SWITCHCAPVCC, 0x3C))
  {
    Serial.println("Initialization failure of SSD1306");
  }
  display.clearDisplay();
  display.setFont(&tahoma7pt7b);
  display.setTextColor(SSD1306_WHITE);
  display.setCursor(0, 15);
  display.println("Extended BLE");
  display.setCursor(0, 31);
  display.println("    Range Test");
  display.setCursor(0, 47);
  display.println("        Peripheral");
  display.setCursor(0, 63);
  display.print("Ver   : ");
  display.println(versionNumber);
  display.display();
  delay(5000);
  display.clearDisplay();

  // Initialization of Bruefruit class
  Bluefruit.configPrphBandwidth(BANDWIDTH_MAX);
  Bluefruit.configUuid128Count(15);
  Bluefruit.begin(1, 0);  // peripheral
  Bluefruit.setTxPower(txpower);
//  Bluefruit.setConnLedInterval(50);
  Bluefruit.autoConnLed(false);   // BLE not used Blue LED
  Bluefruit.Periph.setConnectCallback(connect_callback);
  Bluefruit.Periph.setDisconnectCallback(disconnect_callback);

  // Custom Service and Characteristic Settings
  P2C2P_Service.begin();

//  P2C_Characteristic.setProperties(CHR_PROPS_INDICATE);
  P2C_Characteristic.setProperties(CHR_PROPS_NOTIFY);
  P2C_Characteristic.setFixedLen(DATANUM);  
  P2C_Characteristic.begin();

  C2P_Characteristic.setProperties(CHR_PROPS_READ | CHR_PROPS_WRITE);
  C2P_Characteristic.setFixedLen(DATANUM);
  C2P_Characteristic.setWriteCallback(data_written_callback);  
  C2P_Characteristic.begin();

  // Advertisement Settings
  Bluefruit.Advertising.addFlags(BLE_GAP_ADV_FLAGS_LE_ONLY_GENERAL_DISC_MODE);
  Bluefruit.Advertising.addTxPower();
  Bluefruit.Advertising.addService(P2C2P_Service);
  Bluefruit.ScanResponse.addName(); 
  Bluefruit.Advertising.restartOnDisconnect(true);
  Bluefruit.Advertising.setIntervalMS(20, 153);     // fast mode 20mS, slow mode 153mS
  Bluefruit.Advertising.setFastTimeout(30);         // fast mode 30 sec
  Bluefruit.Advertising.start(0);                   // 0 = Don't stop advertising after n seconds

  Serial.println("End of initialization");

  // Connecting to Central
  Serial.println("Connecting to Central ................");
  while(!Bluefruit.connected()) {
    Serial.print("."); delay(100);
  }
  Serial.println();
  Serial.println("Connected to Central ................");      
}

//*************************************************************************************************************
void loop()
{
  while(connectFlag != true) {
    Serial.print(","); delay(100);
  }

  if(connectFlag == true) 
  {
    digitalWrite(LED_BLUE, LOW);

    // transmit a packet to central ----------------------------------------------
    digitalWrite(LED_RED, LOW);

      // fill buffer with random number
      uint32_t randomnumber = random(0x7fffffff);
      for(int i = 0; i < DATANUM / 4; i++) {
        ud.dataBuff32[i] = randomnumber;
      }
      ud.dataBuff32[0] = millis();  // test start time

//      P2C_Characteristic.indicate(ud.dataBuff8, sizeof(ud.dataBuff8));
//      delay(CONNINTERVAL * 3);  // for Indicate
      P2C_Characteristic.notify(ud.dataBuff8, sizeof(ud.dataBuff8));
      delay(CONNINTERVAL * 6);  // for Notify
    digitalWrite(LED_RED, HIGH);

    // receive a packet from central ---------------------------------------------
    if(writtenFlag == true) 
    {
      digitalWrite(LED_GREEN, LOW);

        memcpy(ud.dataBuff8, readbuff, sizeof(readbuff));
        writtenFlag = false;

        // data received from central
        uint32_t ms = ud.dataBuff32[0];             // test start time
        uint32_t verifynumber = ud.dataBuff32[1];   // random number received by central
        float ave_rssi = ud.dataBuffFL[2];          // average rssi of central
        uint32_t dataLength = ud.dataBuff32[3];     // data length of packet
        int8_t txpower = (int8_t)ud.dataBuff8[16];  // Tx power
        uint8_t phy = ud.dataBuff8[17];             // phy, 1:1Mbps 2:2Mbps 4:coded

        Serial.println("******************************************");
        Serial.print("mS = "); Serial.println(ms);
        Serial.print("randomnumber = "); Serial.println(randomnumber, HEX);      
        Serial.print("verifynumber = "); Serial.println(verifynumber, HEX);
        Serial.print("ave_rssi = "); Serial.println(ave_rssi);      
        Serial.print("dataLength = "); Serial.println(dataLength);
        Serial.print("txpower = "); Serial.println(txpower);      
        Serial.print("phy = "); Serial.println(phy);
        Serial.println();

        display.clearDisplay();
        display.setCursor(80, 15); display.print(verifynumber == randomnumber ? "good" : "bad");
        display.setCursor(0, 15); display.print(randomnumber, HEX);
        display.setCursor(0, 31); display.print(verifynumber, HEX);
        display.setCursor(80, 31);display.print(ms / 1000); display.print(" s");
        display.setCursor(0, 47); display.print(dataLength); display.print(" byte");
        display.setCursor(64, 47);
          if      (phy == 1) { display.print("1 Mbps"); } 
          else if (phy == 2) { display.print("2 Mbps"); } 
          else if (phy == 4) { display.print("Coded"); } 
          else               { display.print("error"); }
        display.setCursor(0, 63); display.print(txpower); display.print(" dBm");
        display.setCursor(64, 63); display.print(ave_rssi, 1); display.print(" dBm");
        display.display();    

        delay(10);    // for LED
      digitalWrite(LED_GREEN, HIGH);
    } // writtenFlag  
    delay(INTERVAL);  // Light Sleep
  } // connectFlag
      digitalWrite(LED_RED, HIGH);    
      digitalWrite(LED_GREEN, HIGH);    
      digitalWrite(LED_BLUE, HIGH);            
} // loop

//**************************************************************************************************************
// callback function
//*******************************************************************************************************
void connect_callback(uint16_t conn_handle)
{
  (void) conn_handle;

  Serial.print("【connect_callback】 conn_Handle : ");
  Serial.println(conn_handle, HEX);

  // Get the reference to current connection
  BLEConnection* connection = Bluefruit.Connection(conn_handle);

  // request PHY changed
  connection->requestPHY(phy); delay(1000);
  // request to update data length    
  connection->requestDataLengthUpdate(); delay(1000);    
  // request MTU exchange
  connection->requestMtuExchange(MTU); delay(1000);
  // request connection interval
  connection->requestConnectionParameter(CONNINTERVAL); delay(1000);

  Serial.println("  Extended BLE settings");
  Serial.print("    PHY ----------> "); Serial.println(connection->getPHY());
  Serial.print("    Data length --> "); Serial.println(connection->getDataLength());
  Serial.print("    MTU ----------> "); Serial.println(connection->getMtu());
  Serial.print("    Interval -----> "); Serial.println(connection->getConnectionInterval());

  connectFlag = true;   
}

//***********************************************************************************************
void disconnect_callback(uint16_t conn_handle, uint8_t reason)
{
  (void) conn_handle;
  (void) reason;

  Serial.print("【disconnect_callback】 reason = 0x");
  Serial.println(reason, HEX);
  connectFlag = false;
}

//***************************************************************************************************
void data_written_callback(uint16_t conn_hdl, BLECharacteristic* chr, uint8_t* data, uint16_t len)
{
  (void) conn_hdl; 
  (void) chr; 
  (void) data; 
  (void) len;

  Serial.print("【written_callback】 len = "); Serial.println(len);
  memcpy(readbuff, data, len);
  writtenFlag = true;
}
